import datetime
import socket
import struct
import logging
from random import randint
from .utils import device_cmd, datetime_to_bytes, datetime_from_bytes, split_list
from .network_utils import packet, DataBuffer
from . import models, debug


class ClientConnection(object):
    def __init__(self, host: str, port: int):
        self._serial = None
        self._secret_key = None
        self._host = host
        self._port = port
        self._socket = None
        self._request = None
        self._response = None
        self._logger = logging.getLogger("buffer") if debug else None

    def disconnect(self):
        if self._socket:
            self._socket.close()
            self._socket = None

    def connect(self):
        self.disconnect()
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.settimeout(20)
        self._socket.connect((self._host, self._port))
        return self._socket is not None

    def send(self):
        if self._request:
            data = bytes(self._request)
            if debug:
                self._logger.debug("REQ ({} bytes): {}".format(len(data),
                                                               str.join(' ', [hex(char) for char in data])))
            self._socket.sendall(data)

    def receive(self, verify_checksum=True, verify_size=False):
        recv = b''
        buffer_size = 4096
        header = self._socket.recv(8)
        if not header:
            raise Exception("connection is closed by server")
        temp = packet(header)
        while len(recv) < temp.size:
            rem_size = temp.size - len(recv)
            temp_data = self._socket.recv(buffer_size if rem_size > buffer_size else rem_size)
            if not temp_data:
                raise Exception("connection is closed by server")
            recv += temp_data
        data = header + recv
        if debug:
            self._logger.debug("RES ({} bytes): {}".format(len(data),
                                                           str.join(' ', [hex(char) for char in data])))
        self._response = packet(data)
        if not self._response.is_valid(verify_checksum, verify_size):
            raise Exception("invalid packet response")

    def verify_response(self, cmd_name: list = None):
        if cmd_name is None:
            cmd_name = ["ack"]
        if self._response.cmd not in [device_cmd[x] for x in cmd_name]:
            raise Exception("unexpected response cmd: {}".format(self._response.cmd))
        # if isinstance(cmd_name, list):
        # else:
        #     if self._response.cmd != device_cmd[cmd_name]:
        #         raise Exception("unexpected response cmd: {}".format(self._response.cmd))

    def send_cmd(self, cmd_name: str, payload: bytes = b'', res_cmd_name: list = None):
        self._request = packet(cmd=device_cmd[cmd_name], serial=self.serial, secret_key=self._secret_key,
                               payload=payload)
        self.send()
        self.receive()
        self.verify_response(res_cmd_name)


class ZMM100_TFT(ClientConnection):
    def __init__(self, host: str, port: int, encoding: str, secret_key: int = 0, serial: int = 0):
        super().__init__(host, port)
        self._encoding = encoding
        self._connected = (self._socket is not None)
        self._secret_key: int = secret_key
        self._serial: int = serial
        self._enabled: bool = True

    def connect(self, comm_key=0):
        # if self._socket:
        #     self.disconnect()
        super().connect()
        self._connected = False
        self._secret_key = 0
        self._serial = 0
        self.send_cmd("connect")
        self._secret_key = self._response.secret_key
        if comm_key > 0:
            payload = self.hash_commkey(comm_key)
            self.send_cmd("login", payload)
        self._connected = True
        return self._connected

    def disconnect(self):
        try:
            if self._connected:
                self.send_cmd("disconnect")
        except Exception as ex:
            if self._logger:
                self._logger.error("disconnection with error:")
                self._logger.error(str(ex))
        finally:
            super().disconnect()
            self._connected = False

    def hash_commkey(self, comm_key: int):
        index = 1
        num = 0
        for _ in range(0x20):
            num *= 2
            num = num & 0xFFFFFFFF
            if comm_key & index > 0:
                num = num | 1
            index = index << 1
            if index > 0xFFFFFFFF:
                index = 1
        num += self._secret_key
        # y1 = (num & 0xFF) ^ 0x5A
        y2 = (num >> 8 & 0xFF) ^ 0x4B
        y3 = (num >> 16 & 0xFF) ^ 0x53
        y4 = (num >> 24 & 0xFF) ^ 0x4F
        y5 = randint(1, 0x100)
        y2 = y2 ^ y5
        y3 = y3 ^ y5
        y4 = y4 ^ y5
        ret = [y2, y5, y4, y3]
        ret.reverse()
        return bytes(ret)

    def reboot(self):
        self.send_cmd("reboot")

    def shutdown(self):
        self.send_cmd("shutdown")

    def set_device_prop(self, new_value: bytes, prop_cmd="set_data"):
        if not self._connected:
            raise Exception("you need to connect to machine first")
        self.send_cmd(prop_cmd, new_value)
        self.send_cmd("save_data")
        # return True

    @property
    def id(self):
        return self.get_device_prop(b"DeviceID\x00")

    @id.setter
    def id(self, val):
        payload = "DeviceID={}".format(val)
        self.set_device_prop(payload.encode(self._encoding))

    @property
    def name(self):
        return self.get_device_prop(b"~DeviceName\x00")

    @property
    def type(self):
        return self.get_device_prop(b"DeviceType\x00")

    @property
    def product_time(self):
        return self.get_device_prop(b"~ProductTime")

    @property
    def serial_number(self):
        return self.get_device_prop(b"~SerialNumber\x00")

    @property
    def language(self):
        return self.get_device_prop(b"Language\x00")

    @property
    def compat_old_firmware(self):
        return self.get_device_prop(b"CompatOldFirmware\x00")

    @property
    def is_support_pull(self):
        return self.get_device_prop(b"IsSupportPull\x00")

    @property
    def camera_open(self):
        return self.get_device_prop(b"CameraOpen\x00")

    @property
    def finger_fun_on(self):
        return self.get_device_prop(b"FingerFunOn")

    @property
    def face_fun_on(self):
        return self.get_device_prop(b"FaceFunOn\x00")

    @property
    def zk_face_version(self):
        return self.get_device_prop(b"ZKFaceVersion")

    @property
    def biometric_type(self):
        return self.get_device_prop(b"BiometricType")

    @property
    def build_version(self):
        return self.get_device_prop(b"BuildVersion\x00")

    @property
    def att_photo_for_sdk(self):
        return self.get_device_prop(b"AttPhotoForSDK\x00")

    @property
    def is_only_rf_machine(self):
        return self.get_device_prop(b"~IsOnlyRFMachine\x00")

    @property
    def ssr(self):
        return self.get_device_prop(b"~SSR\x00")

    @property
    def pin_width(self):
        return self.get_device_prop(b"~PIN2Width\x00")

    @property
    def vendor(self):
        return self.get_device_prop(b"~OEMVendor\x00")

    @property
    def platform(self):
        return self.get_device_prop(b"~Platform\x00")

    @property
    def os(self):
        return self.get_device_prop(b"~OS")

    @property
    def extend_fmt_1(self):
        return self.get_device_prop(b"~ExtendFmt")

    @property
    def extend_fmt_2(self):
        return self.get_device_prop(b"ExtendFmt\x00")

    @property
    def extend_oplog_1(self):
        return self.get_device_prop(b"~ExtendOPLog\x00")

    @property
    def user_ext_fmt(self):
        return self.get_device_prop(b"~UserExtFmt\x00")

    @property
    def extend_oplog_2(self):
        return self.get_device_prop(b"ExtendOPLog")

    @property
    def software_version(self):
        if not self._connected:
            raise Exception("you need to connect to machine first")
        self.send_cmd("soft_ver")
        msg = self._response.payload.decode('latin-1')
        return msg.strip(' \x00')

    @property
    def fp_version(self):
        return self.get_device_prop(b"~ZKFPVersion\x00")

    @property
    def work_code(self):
        return self.get_device_prop(b"WorkCode\x00")

    @work_code.setter
    def work_code(self, val: str):
        temp = "WorkCode=" + val
        self.set_device_prop(temp.encode('latin-1'))

    @property
    def mac_address(self):
        return self.get_device_prop(b"MAC\x00")

    @mac_address.setter
    def mac_address(self, val: str):
        temp = "MAC=" + val
        self.set_device_prop(temp.encode('latin-1'))

    @property
    def ip_address(self):
        return self.get_device_prop(b"IPAddress\x00")

    @ip_address.setter
    def ip_address(self, val: str):
        temp = "IPAddress=" + val
        self.set_device_prop(temp.encode('latin-1'))

    @property
    def password(self):
        return self.get_device_prop(b"COMKey\x00")

    @password.setter
    def password(self, val):
        payload = "COMKey={}".format(val)
        self.set_device_prop(payload.encode('latin-1'))

    @property
    def port(self):
        return int(self.get_device_prop(b"UDPPort\x00"))

    @port.setter
    def port(self, val: int):
        payload = "UDPPort={}".format(val)
        self.set_device_prop(payload.encode('latin-1'))

    @property
    def daylight_saving_timeon(self):
        return self.get_device_prop(b"DaylightSavingTimeOn\x00")

    @daylight_saving_timeon.setter
    def daylight_saving_timeon(self, val: str):
        payload = "DaylightSavingTimeOn=" + val
        self.set_device_prop(payload.encode('latin-1'))

    @property
    def daylight_saving_time(self):
        return self.get_device_prop(b"DaylightSavingTime\x00")

    @daylight_saving_time.setter
    def daylight_saving_time(self, val: str):
        payload = "DaylightSavingTime=" + val
        self.set_device_prop(payload.encode('latin-1'))

    @property
    def standard_time(self):
        return self.get_device_prop(b"StandardTime\x00")

    @standard_time.setter
    def standard_time(self, val: str):
        payload = "StandardTime=" + val
        self.set_device_prop(payload.encode('latin-1'))

    @property
    def device_time(self):
        if not self._connected:
            raise Exception("you need to connect to machine first")
        self.send_cmd("get_time")
        return datetime_from_bytes(self._response.payload[:4])

    @device_time.setter
    def device_time(self, new_time: datetime.datetime):
        self.set_device_prop(datetime_to_bytes(new_time), "set_time")

    def query_system_option(self,
                            query_string="~OS=?,ExtendFmt=?,~ExtendFmt=?,ExtendOPLog=?,~ExtendOPLog=?,~Platform=?,"
                                         "~ZKFPVersion=?,WorkCode=?,~SSR=?,~PIN2Width=?,~UserExtFmt=?,BuildVersion=?,"
                                         "AttPhotoForSDK=?,~IsOnlyRFMachine=?,CameraOpen=?,CompatOldFirmware=?,"
                                         "IsSupportPull=?,Language=?,~SerialNumber=?,FaceFunOn=?,~DeviceName=?"):
        self.send_cmd("query_sys_op", query_string.encode('latin-1'))
        return self._response.payload.decode("latin-1")

    def get_device_prop(self, prop_payload: bytes):
        if not self._connected:
            raise Exception("you need to connect to machine first")
        self.send_cmd("get_data", prop_payload, ["ack", "no_sys_op"])
        if self._response.cmd == device_cmd["no_sys_op"]:
            return None
        msg = self._response.payload.decode('latin-1')
        return msg.split("=")[1].strip(' \x00')

    def get_tz_info(self, tz_index: int):
        if not self._connected:
            raise Exception("you need to connect to machine first")

        self.send_cmd("get_tz", struct.pack("<I", tz_index))
        index = struct.unpack("<H", self._response.payload[:2])[0]
        if index == tz_index:
            tz = self._response.payload[2:-2]
            return ''.join(['{0:02d}'.format(x) for x in tz])

    def set_tz_info(self, tz_index: int, str_val):
        if not self._connected:
            raise Exception("you need to connect to machine first")
        if len(str_val) % 2 > 0:
            return False
        vals = split_list(str_val, len(str_val) // 2)
        payload = struct.pack("<H", tz_index) + bytes([int(x) for x in vals])
        self.send_cmd("set_tz", payload)
        return True

    def clear_data(self, data_id: int):
        if not self._connected:
            raise Exception("you need to connect to machine first")
        self.send_cmd("cls_data", payload=struct.pack("<B", data_id))
        self.send_cmd("save_data")

    def del_user(self, _id: int, disable_device=True):
        if not self._connected:
            raise Exception("you need to connect to machine first")
        if disable_device:
            self.disable_device()
        try:
            self.send_cmd("del_user", payload=struct.pack("<H", _id))
        except Exception as ex:
            raise ex
        finally:
            if disable_device:
                self.enable_device()

    @property
    def database_structure(self):
        if not self._connected:
            raise Exception("you need to connect to machine first")
        self.send_cmd("get_table_struct", res_cmd_name=["recv_buff_header"])
        self.receive()
        self.verify_response(["recv_buff_content"])
        data = self._response
        self.receive()
        self.verify_response()
        # footer = self._response
        return data.payload.decode('latin-1')

    def get_fp_data(self, user_id: int, finger_id: int, disable_device=True):
        """
        return byte data of the fingerprint
        """
        if not self._connected:
            raise Exception("you need to connect to machine first")
        if disable_device:
            self.disable_device()
        try:
            self.send_cmd("get_fp", struct.pack("<HB", user_id, finger_id), ["recv_buff_header", "nak", "no_data"])
            if self._response.cmd == device_cmd["nak"]:
                return b''
            elif self._response.cmd == device_cmd["no_data"]:
                return b''
            else:
                self.receive()
                self.verify_response(["recv_buff_content"])
                data = self._response
                self.receive()
                self.verify_response()
                # footer = self._response
                return data.payload[:-6]
        except Exception as ex:
            raise ex
        finally:
            if disable_device:
                self.enable_device()

    def send_file(self, file_name: str, file_data: bytes, disable_device=True):
        if not self._connected:
            raise Exception("you need to connect to machine first")
        if disable_device:
            self.disable_device()
        try:
            self._upload_data(file_data)
            file_name = file_name.encode(self._encoding)
            file_name = file_name + struct.pack("<{}x".format(39 - len(file_name)))
            payload = struct.pack("<I", 0x6A4) + file_name
            self.send_cmd("send_file", payload, ["ack", "nak"])
            ret = self._response.cmd == device_cmd["ack"]
            self.send_cmd("end_buff_stream")
            if not ret:
                raise Exception("operation failed")
        finally:
            if disable_device:
                self.enable_device()
        self.send_cmd("save_data")

    def set_user_pic(self, person_id: str, pic_data: bytes, disable_device=True):
        if not self._connected:
            raise Exception("you need to connect to machine first")
        if disable_device:
            self.disable_device()
        try:
            self._upload_data(pic_data)
            file_name = person_id.encode(self._encoding) + b".jpg"
            file_name = file_name + struct.pack("<{}x".format(39 - len(file_name)))
            payload = file_name + bytes([0] * 4)
            self.send_cmd("set_user_pic", payload)
            self.send_cmd("end_buff_stream")
        finally:
            if disable_device:
                self.enable_device()
        self.send_cmd("save_data")

    def del_user_pic(self, person_id: str, disable_device=True):
        if not self._connected:
            raise Exception("you need to connect to machine first")
        if disable_device:
            self.disable_device()
        try:
            self.send_cmd("del_user_pic", person_id.encode(self._encoding) + b".jpg\x00", ["ack", "no_pic"])
            if self._response.cmd == device_cmd["no_pic"]:
                return False
            return True
        finally:
            if disable_device:
                self.enable_device()
            self.send_cmd("save_data")

    def get_user_pic(self, person_id: str):
        if not self._connected:
            raise Exception("you need to connect to machine first")
        self.send_cmd("get_user_pic", person_id.encode(self._encoding) + b".jpg\x00",
                      ["recv_buff_header", "nak", "no_pic"])
        if self._response.cmd != device_cmd["recv_buff_header"]:
            return b''
        data_size, _ = struct.unpack("<II", self._response.payload[:8])
        self.receive()
        self.verify_response(["recv_buff_content"])
        data = self._response
        self.receive()
        self.verify_response()
        _ = self._response
        if len(data.payload) != data_size:
            raise Exception("data integrity is incorrect")
        return data.payload

    def del_op_logs(self):
        self.send_cmd("clear_op_log")
        self.send_cmd("save_data")

    def del_users(self):
        self.clear_data(0x05)

    def del_fps(self):
        self.clear_data(0x02)

    def del_admins(self):
        if not self._connected:
            raise Exception("you need to connect to machine first")
        self.send_cmd("cls_admins")
        self.send_cmd("save_data")

    def get_photo_count(self):
        if not self._connected:
            raise Exception("you need to connect to machine first")
        self.send_cmd("get_photo_count")
        if len(self._response.payload) == 1:
            return struct.unpack("<B", self._response.payload)[0]
        elif len(self._response.payload) == 2:
            return struct.unpack("<H", self._response.payload)[0]
        elif len(self._response.payload) == 4:
            return struct.unpack("<I", self._response.payload)[0]
        else:
            raise Exception("unknown response format for photo count")

    def get_state(self, disable_device=True):
        if not self._connected:
            raise Exception("you need to connect to machine first")
        if disable_device:
            self.disable_device()
        try:
            self.send_cmd("logs_count")

            if len(self._response.payload or b'') < (23 * 4):
                raise Exception("error in logs_count command")
            return models.MachineState.from_bytes(self._response.payload)
        except Exception as ex:
            raise ex
        finally:
            if disable_device:
                self.enable_device()

    def del_att_logs(self, disable_device=True):
        if not self._connected:
            raise Exception("you need to connect to machine first")
        if disable_device:
            self.disable_device()
        try:
            self.send_cmd("del_logs")
        except Exception as ex:
            raise ex
        finally:
            if disable_device:
                self.enable_device()
        self.send_cmd("save_data")

    def disable_device(self, time_out_in_sec=0):
        if not self._connected:
            raise Exception("you need to connect to machine first")
        self.send_cmd("disable", struct.pack("<I", time_out_in_sec))

    def enable_device(self):
        if not self._connected:
            raise Exception("you need to connect to machine first")
        self.send_cmd("enable")

    def check_hash(self):
        """
        use as follows after send data from client to machine using header-content-footer
        use this function optionally to return the hash of the buffer received by device and
        compare it with hash you make from buffer before send
        """
        if not self._connected:
            raise Exception("you need to connect to machine first")
        self.send_cmd("check_hash")
        return struct.unpack("<I", self._response.payload)

    def get_fps(self, disable_device=True):
        ret = self._get_data_buffer(0x02000701, disable_device)
        if ret is not None:
            ret.remove_leading()
            return ret.fps

    def del_user_face(self, person_id: str, face_id=50, disable_device=True):
        # ret = False
        if not self._connected:
            raise Exception("you need to connect to machine first")
        if disable_device:
            self.disable_device()
        try:
            payload = [0] * 28
            temp = person_id.encode(self._encoding)
            length = len(temp)
            if length > 24:
                length = 24
            payload[:length] = temp[:length]
            payload[25] = face_id
            self.send_cmd("del_face", bytes(payload))
        finally:
            if disable_device:
                self.enable_device()
        self.send_cmd("save_data")

    def get_user_face(self, person_id: str, face_id=50, disable_device=True):
        if not self._connected:
            raise Exception("you need to connect to machine first")
        if disable_device:
            self.disable_device()
        try:
            # payload = [0] * 28
            person_id = person_id.encode(self._encoding)
            payload = person_id + struct.pack("<{}xxB2x".format(24 - len(person_id)), face_id)
            self.send_cmd("get_face", payload, ["recv_buff_header", "nak", "no_data"])
            if self._response.cmd == device_cmd["nak"]:
                return b''
            elif self._response.cmd == device_cmd["no_data"]:
                return b''
            else:
                self.receive()
                self.verify_response(["recv_buff_content"])
                data = self._response
                self.receive()
                self.verify_response()
                _ = self._response
                return data.payload
        finally:
            if disable_device:
                self.enable_device()

    def set_user_face(self, person_id: str, face_data: bytes, face_index=50, disable_device=True):
        if not self._connected:
            raise Exception("you need to connect to machine first")
        if disable_device:
            self.disable_device()
        try:
            self._upload_data(face_data)
            person_id = person_id.encode(self._encoding)
            payload = person_id + struct.pack("<{}xxBH".format(24 - len(person_id)), face_index, len(face_data))
            self.send_cmd("set_face", payload)
            self.send_cmd("end_buff_stream")
        finally:
            if disable_device:
                self.enable_device()
        self.send_cmd("save_data")

    # old function
    # def delFP(self, user_id, finger_id, disable_device=True):
    #     if not self._connected:
    #         raise Exception("you need to connect to machine first")
    #     if disable_device:
    #         self.disableDevice()
    #     try:
    #         self.sendCmd("del_fp", struct.pack("<HB", user_id, finger_id), ["ack", "nak", "no_fp"])
    #         ret = self._response.cmd == device_cmd["ack"]
    #         self.sendCmd("save_data")
    #         return ret
    #     finally:
    #         if disable_device:
    #             self.enableDevice()
    def set_user(self, user_info: models.UserInfo):
        """
        It is very important to identify the serial of the user_info object
        because the machine use this serial to store data thus avoid overriding
        on user over another one thus please use correct index (serial)
        """
        if not self._connected:
            raise Exception("you need to connect to machine first")
        self.send_cmd("set_user", bytes(user_info))

    def del_fp(self, person_id: str, finger_id: int, disable_device=True):
        # ret = False
        if not self._connected:
            raise Exception("you need to connect to machine first")
        if disable_device:
            self.disable_device()
        try:
            payload = [0] * 25
            temp = person_id.encode(self._encoding)
            length = len(temp)
            if length > 24:
                length = 24
            payload[:length] = temp[:length]
            payload[24] = finger_id
            # if len(temp) >= 24:
            #     payload = temp[:24] + struct.pack("<B", finger_id)
            # else:
            #     remain = 24 - len(temp)
            #     fmt = "<{}xB".format(remain)
            #     payload = temp + struct.pack(fmt, finger_id)
            self.send_cmd("del_fp_ex", bytes(payload), ["ack", "nak", "no_fp"])
            ret = self._response.cmd == device_cmd["ack"]
        finally:
            if disable_device:
                self.enable_device()
        if ret:
            self.send_cmd("save_data")
        return ret

    def _upload_data(self, data: bytes):
        self.send_cmd("recv_buff_header", struct.pack("<I", len(data)))
        self.send_cmd("recv_buff_content", data)
        buffer = DataBuffer(self._encoding, data)
        hash_code = hash(buffer)
        self.send_cmd("check_hash")
        if hash_code != struct.unpack("<I", self._response.payload)[0]:
            raise Exception("problem in data sending")

    def set_fp(self, fp_info: models.FPInfo, disable_device=True):
        if not self._connected:
            raise Exception("you need to connect to machine first")
        if disable_device:
            self.disable_device()
        try:
            self._upload_data(fp_info.data)
            # self.sendCmd("recv_buff_header", struct.pack("<I", len(fp_info.data)))
            # self.sendCmd("recv_buff_content", fp_info.data)
            # buffer = DataBuffer(fp_info.data)
            # hash_code = hash(buffer)
            # self.sendCmd("check_hash")
            # if hash_code != struct.unpack("<I", self._response.payload)[0]:
            #     raise Exception("problem in data sending")
            payload = struct.pack("<HBBH", fp_info.user_id, fp_info.finger_id, fp_info.enabled, len(fp_info.data))
            self.send_cmd("set_fp_ex", payload)
            self.send_cmd("end_buff_stream")
        finally:
            if disable_device:
                self.enable_device()
        self.send_cmd("save_data")

    def set_fps(self, fp_infos: list, disable_device=True):
        if not self._connected:
            raise Exception("you need to connect to machine first")
        if disable_device:
            self.disable_device()
        try:
            for fp_info in fp_infos:
                self.send_cmd("recv_buff_header", struct.pack("<I", len(fp_info.data)))
                self.send_cmd("recv_buff_content", fp_info.data)
                buffer = DataBuffer(self._encoding, fp_info.data)
                hash_code = hash(buffer)
                self.send_cmd("check_hash")
                if hash_code != struct.unpack("<I", self._response.payload)[0]:
                    raise Exception("problem in data sending")
                payload = struct.pack("<HBBH", fp_info.user_id, fp_info.finger_id, fp_info.enabled, len(fp_info.data))
                self.send_cmd("set_fp_ex", payload)
                self.send_cmd("end_buff_stream")
        finally:
            if disable_device:
                self.enable_device()
        self.send_cmd("save_data")

    def get_fp(self, user_id: int, finger_id: int, disable_device=True):
        """
        return FPInfo record
        """
        if not self._connected:
            raise Exception("you need to connect to machine first")
        if disable_device:
            self.disable_device()
        try:
            self.send_cmd("get_fp_ex", struct.pack("<HB", user_id, finger_id), ["recv_buff_header", "nak", "no_data"])
            if self._response.cmd == device_cmd["nak"]:
                return
            elif self._response.cmd == device_cmd["no_data"]:
                return
            else:
                self.receive()
                self.verify_response(["recv_buff_content"])
                data = self._response
                self.receive()
                self.verify_response()
                # footer = self._response
                return models.FPInfo(user_id=user_id, finger_id=finger_id, enabled=data.payload[-1],
                                     data=data.payload[:-7])
        except Exception as ex:
            raise ex
        finally:
            if disable_device:
                self.enable_device()

    def get_users(self, disable_device=True):
        ret = self._get_data_buffer(0x05000901, disable_device)
        if ret is not None:
            ret.remove_leading()
            return ret.users

    def get_op_logs(self, disable_device=True):
        ret = self._get_data_buffer(0x2201, disable_device)
        if ret is not None:
            ret.remove_leading()
            return ret.op_logs

    def get_att_logs(self, disable_device=True):
        ret = self._get_data_buffer(0x0D01, disable_device)
        if ret is not None:
            ret.remove_leading()
            return ret.att_logs

    def _get_data_buffer(self, data_id: int, disable_device=True):
        if not self._connected:
            raise Exception("you need to connect to machine first")
        if disable_device:
            self.disable_device()
        try:

            data_buffer = DataBuffer(self._encoding)
            payload = struct.pack("<I7x", data_id)
            self.send_cmd("start_buff_stream", payload, ["ack", "nak", "recv_buff_content", "no_record"])
            if self._response.cmd == device_cmd["recv_buff_content"]:
                data_buffer.append(self._response.payload)
                return data_buffer
            elif self._response.cmd in [device_cmd["nak"], device_cmd["no_record"]]:
                return data_buffer
            else:
                data_size = struct.unpack("<I", self._response.payload[1:5])[0]
                packet_size = 0xFFC0
                packets = data_size // packet_size
                rem_size = data_size
                if (data_size % packet_size) > 0:
                    packets += 1
                # has_error = False
                for _ in range(0, packets):
                    # payload = DwordToBytes(data_size - rem_size) + DwordToBytes(rem_size if rem_size < packet_size
                    # else packet_size)
                    payload = struct.pack("<I", data_size - rem_size) + \
                              struct.pack("<I", rem_size if rem_size < packet_size else packet_size)
                    self.send_cmd("buff_stream", payload, ["recv_buff_header"])
                    recv_size = struct.unpack("<I", self._response.payload[:4])[0]
                    self.receive()
                    self.verify_response(["recv_buff_content"])
                    data = self._response
                    self.receive()
                    self.verify_response()
                    # footer = self._response
                    data_buffer.append(data.payload)
                    rem_size -= recv_size
                self.send_cmd("end_buff_stream")
                # if not has_error:
                return data_buffer
        except Exception as ex:
            raise ex
        finally:
            if disable_device:
                self.enable_device()

    @property
    def connected(self):
        return self._connected

    @property
    def serial(self):
        self._serial += 1
        return self._serial - 1


class ZMM220_TFT(ZMM100_TFT):
    def __init__(self, host, port, encoding, secret_key=0, serial=0):
        super().__init__(host, port, encoding, secret_key, serial)

    def connect(self, comm_key=0):
        if self._socket:
            self.disconnect()
        super(ZMM100_TFT, self).connect()
        self._connected = False
        self._secret_key = 0
        self._serial = 0
        self.send_cmd("connect", res_cmd_name=["accept_conn"])
        self._secret_key = self._response.secret_key
        if comm_key > 0:
            payload = self.hash_commkey(comm_key)
            self.send_cmd("login", payload)
        self._connected = True
        return self._connected



from datetime import datetime
import struct
# if segment size > 0x2000 shift to stream
device_cmd = {
    'read_db': 0x0007,
    'set_user': 0x0008,
    'get_fp': 0x0009,
    'set_fp': 0x000A,
    'get_data': 0x000B,  # another name "get_sys_option"
    'set_data': 0x000C,  # another name "set_sys_option"
    'read_latest_log': 0x000D,
    'cls_data': 0x000E,
    'del_logs': 0x000F,
    'append_fp': 0x0011,
    'del_user': 0x0012,
    'del_fp': 0x0013,
    'cls_admins': 0x0014,
    'get_user_group': 0x0015,
    'set_user_group': 0x0016,
    'get_user_tz': 0x0017,
    'set_user_tz': 0x0018,
    'get_group_tz': 0x0019,
    'set_group_tz': 0x001A,
    'get_tz': 0x001B,
    'set_tz': 0x001C,
    'get_unlock_group': 0x001D,
    "set_lock_comb": 0x001E,
    "unlock": 0x001F,
    'clear_op_log': 0x0021,
    'read_op_log': 0x0022,
    'ssr_read_att_role': 0x0023,
    'ssr_read_depart': 0x0024,
    'ssr_read_turn': 0x00025,
    'logs_count': 0x0032,
    'end_enroll': 0x003C,
    'start_enroll': 0x003D,
    'cancel_op': 0x003E,
    'query_state': 0x0040,
    'write_LCD': 0x0042,
    'clear_LCD': 0x0043,
    'get_pin_width': 0x0045,
    'get_sms_id': 0x0047,
    'set_user_sms_id': 0x0049,
    'get_door_state': 0x004B,
    'set_workcode': 0x0052,
    'get_workcode': 0x0053,
    'set_fp_ex': 0x0057,
    'get_fp_ex': 0x0058,
    'read_rt_log': 0x005A,
    'get_holiday_tz': 0x005B,
    "set_holiday_tz": 0x005C,
    "send_file": 0x006E,
    'check_hash': 0x0077, 
    'ssr_del_user_ext': 0x0085,
    'del_fp_ex': 0x0086,
    'del_user_fp': 0x0086,
    'get_face': 0x0096,
    "set_face": 0x0097,
    "del_face": 0x0098,
    'get_time': 0x00C9,
    'set_time': 0x00CA,
    'get_holiday': 0x012C,
    'set_holiday': 0x012D,
    'reg_event': 0x01F4,
    'query_sys_op': 0x1F5,
    'connect': 0x03E8,
    'disconnect': 0x03E9,
    'enable': 0x03EA,
    'disable': 0x03EB,
    'reboot': 0x03EC,
    'shutdown': 0x03ED,
    "sleep": 0x03EE,
    'resume': 0x03EF,
    'capture_image': 0x03F4,
    'save_data': 0x03F5,
    'play_voice': 0x03F9,
    'beep': 0x03FC,
    'suspend': 0x03FE,
    'soft_ver': 0x044C,
    'login': 0x044E,
    'recv_buff_header': 0x05DC,
    'recv_buff_content': 0x05DD,
    'end_buff_stream': 0x05DE,
    'start_buff_stream': 0x05DF,
    'buff_stream': 0x05E0,
    'update_lang_by_id': 0x06A5,
    'set_custom_att_state': 0x06A7,
    "set_custom_voice": 0x06A8,
    'enable_custom_voice': 0x06AA,
    "update_file": 0x06AD,
    'ack': 0x07D0,
    'nak': 0x07D1,
    'accept_conn': 0x07D5,
    'get_photo_count': 0x07DD,
    'get_photo_by_name': 0x07DE,
    'clear_photo_by_time': 0x07DF,
    'get_photo_names_by_time': 0x07E0,
    'read_large_tmp': 0x07E1,
    'read_temp_data': 0x08AD,
    'no_pic': 0x1379,
    'no_record': 0x137D,
    'no_fp': 0x137F,
    'no_data': 0x1381,
    'empty_upload_buffer': 0x1383,  # received in set_fp_ex command
    'no_sys_op': 0x1387,
    'get_table_struct': 0x2710,
    'set_user_pic': 0x2719,
    'get_user_pic': 0x271A,
    'del_user_pic': 0x271B,
    'not_support': 0xFFFF
}


def dump(data):
    return [hex(x) for x in data]


def split_list(alist, wanted_parts=1):
    length = len(alist)
    return [alist[i * length // wanted_parts: (i + 1) * length // wanted_parts] for i in range(wanted_parts)]


def datetime_to_number(dt):
    year = dt.year - 2000
    month = dt.month - 1
    day = dt.day - 1
    hour = dt.hour
    minute = dt.minute
    second = dt.second
    temp = year * 12
    temp = temp + month
    temp = (temp << 5) - temp

    temp = ((temp + day) * 24) + hour
    temp = (temp << 4) - temp

    temp = (temp * 4) + minute
    temp = (temp << 4) - temp

    temp = (temp * 4) + second
    return temp


def number_to_datetime(dword_time):
    base_number = dword_time
    temp1 = ((base_number * 0x88888889) >> 32) & 0xFFFFFFFF
    temp2 = temp1 >> 5
    temp3 = temp2 << 4
    second = base_number - ((temp3 - temp2) * 4)
    base_number = temp2
    temp1 = ((base_number * 0x88888889) >> 32) & 0xFFFFFFFF
    temp2 = temp1 >> 5
    temp3 = temp2 << 4
    minute = base_number - ((temp3 - temp2) * 4)

    base_number = temp2
    temp1 = ((base_number * 0xAAAAAAAB) >> 32) & 0xFFFFFFFF
    temp2 = temp1 >> 4
    temp3 = temp2 * 3
    hour = base_number - (temp3 * 8)

    base_number = temp2
    temp1 = ((base_number * 0x8421085) >> 32) & 0xFFFFFFFF
    temp2 = (((base_number - temp1) >> 1) + temp1) >> 4
    temp3 = temp2 << 5
    day = base_number - (temp3 - temp2) + 1

    base_number = temp1
    temp1 = ((base_number * 0xAAAAAAAB) >> 32) & 0xFFFFFFFF
    temp2 = (temp1 >> 3)
    temp3 = temp2 * 12
    month = (base_number - temp3) + 1
    year = temp2 + 0x7D0
    return datetime(year, month, day, hour, minute, second)


def datetime_to_bytes(dt):
    return struct.pack("<I", datetime_to_number(dt))


def datetime_from_bytes(data):
    return number_to_datetime(struct.unpack("<I", data)[0])


def get_null_term_str(data, start, end, encoding):
    temp = data[start:end]
    if 0 in temp:
        temp = bytes(temp[:temp.index(0)])
    else:
        temp = bytes(temp)
    return temp.decode(encoding)



import struct

from .utils import get_null_term_str, number_to_datetime, datetime_to_number


class UserInfo:
    def __init__(self, encoding):
        self.id = 0
        self.machine_id = 0
        self.name = ""
        self.password = ""
        self.enabled = True
        self.privilege = 0
        self.card_no = 0
        self.person_id = ""
        self.pic = None
        self._data = None
        self._encoding = encoding

    @staticmethod
    def from_bytes(data, encoding):
        model = UserInfo(encoding)
        model._data = data
        if data:
            model.id = struct.unpack("<H", data[:2])[0]
            temp = int(data[2])
            model.enabled = (temp & 1 == 0)
            temp = temp >> 1
            if temp == 1:
                model.privilege = 1
            elif temp == 3:
                model.privilege = 2
            elif temp == 7:
                model.privilege = 3
            else:
                model.privilege = 0
            model.password = get_null_term_str(data, 3, 11, encoding)
            model.name = get_null_term_str(data, 11, 35, encoding)
            model.card_no = struct.unpack("<I", data[35:39])[0]
            model.person_id = get_null_term_str(data, 48, None, encoding)
        return model

    def __bytes__(self):
        output = [0] * 72
        output[0:2] = struct.pack("<H", self.id)
        if self.privilege == 2:
            temp = 3
        elif self.privilege == 3:
            temp = 7
        else:
            temp = self.privilege
        temp = temp << 1
        temp = temp | (0 if self.enabled else 1)
        output[2] = temp
        temp = self.password.encode(self._encoding)
        output[3: 3 + len(temp)] = temp
        temp = self.name.encode(self._encoding)
        output[11: 11 + len(temp)] = temp
        output[35:39] = struct.pack("<I", self.card_no)
        output[39] = 1
        temp = self.person_id.encode(self._encoding)
        output[48: 48 + len(temp)] = temp
        return bytes(output)


class FPInfo:
    def __init__(self, user_id=0, finger_id=0, enabled=True, data=None):
        self.machine_id = 0
        self.user_id = user_id
        self.finger_id = finger_id
        self.enabled = enabled
        self.data = data

    @staticmethod
    def from_bytes(data):
        model = FPInfo()
        if data:
            model.user_id, model.finger_id, model.enabled = struct.unpack("<HBB", data[:4])
            model.data = data[4:]
        return model

    def __bytes__(self):
        return struct.pack("<HBB", self.user_id, self.finger_id, self.enabled) + self.data


class MachineState:
    def __init__(self):
        self.machine_id = 0
        self.user_count = 0
        self.finger_count = 0
        self.face_count = 0
        self.record_count = 0
        self.op_record_count = 0
        self.admin_count = 0
        self.password_count = 0
        self.user_max = 0
        self.finger_max = 0
        self.face_max = 0
        self.record_max = 0
        self.user_rem = 0
        self.finger_rem = 0
        self.face_rem = 0
        self.record_rem = 0

    # machine_id = Column(Integer, ForeignKey(Machine.id), primary_key=True)
    # user_count = Column(Integer)
    # finger_count = Column(Integer)
    # face_count = Column(Integer)
    # record_count = Column(Integer)
    # op_record_count = Column(Integer)
    # admin_count = Column(Integer)
    # password_count = Column(Integer)
    # user_max = Column(Integer)
    # finger_max = Column(Integer)
    # face_max = Column(Integer)
    # record_max = Column(Integer)
    # user_rem = Column(Integer)
    # finger_rem = Column(Integer)
    # face_rem = Column(Integer)
    # record_rem = Column(Integer)

    @staticmethod
    def from_bytes(data):
        model = MachineState()
        model.user_count, \
        model.finger_count, \
        model.record_count, \
        model.op_record_count, \
        model.admin_count, \
        model.password_count, \
        model.finger_max, \
        model.user_max, \
        model.record_max, \
        model.finger_rem, \
        model.user_rem, \
        model.record_rem, \
        model.face_count, \
        model.face_rem, \
        model.face_max = struct.unpack("<16xI4xI4xI4xI4x11I", data[:23 * 4])
        return model

    def __bytes__(self):
        return struct.pack("<16xI4xI4xI4xI4x11I20x",
                           self.user_count,
                           self.finger_count,
                           self.record_count,
                           self.op_record_count,
                           self.admin_count,
                           self.password_count,
                           self.finger_max,
                           self.user_max,
                           self.record_max,
                           self.finger_rem,
                           self.user_rem,
                           self.record_rem,
                           self.face_count,
                           self.face_rem,
                           self.face_max)


class AttLog:
    def __init__(self):
        self.machine_id = 0
        self.att_time = None
        self.person_id = ""
        self.serial = 0
        self.verify_mode = 0
        self.in_out = 0
        self.work_code = 0
        self.encoding = ""

    @staticmethod
    def from_bytes(data, encoding):
        model = AttLog()
        if data:
            model.serial = struct.unpack("<H", data[:2])[0]
            model.person_id = get_null_term_str(data, 2, 26, encoding)
            model.verify_mode, date, model.in_out, model.work_code = struct.unpack("<BIBH", data[26:34])
            model.att_time = number_to_datetime(date)
            model.encoding = encoding
        return model

    def __bytes__(self):
        output = [0] * 40
        output[0:2] = struct.pack("<H", self.serial)
        b_userid = self.person_id.encode(self.encoding)
        output[2:2 + len(b_userid)] = b_userid
        output[26:34] = struct.pack("<BIBH", self.verify_mode, datetime_to_number(self.att_time), self.in_out,
                                    self.work_code)
        return bytes(output)


class OpLog:
    def __init__(self):
        self.machine_id = 0
        self.op_id = 0
        self.op_time = None
        self.admin = 0
        self.param_1 = 0
        self.param_2 = 0
        self.param_3 = 0
        self.param_4 = 0

    @staticmethod
    def from_bytes(data):
        model = OpLog()
        if data:
            model.admin, model.op_id, date, model.param_1, model.param_2, model.param_3, model.param_4 \
                 = struct.unpack("<HHIHHHH", data[:16])
            model.op_time = number_to_datetime(date)
        return model


import struct

from .models import UserInfo, AttLog, FPInfo, OpLog
from .utils import split_list


class packet(object):
    def __init__(self, data=None, cmd=0, serial=0, payload=b"", secret_key=0xC31E):
        self._header = b"\x50\x50\x82\x7D"
        self._cmd = cmd
        self._checksum = 0x00
        self._secret_key = secret_key  # generate random word key
        self._serial = serial
        self._payload = payload
        self._size = len(self._payload) + 8
        if data is not None:
            if len(data) >= 8:
                self._header = data[0:4]
                self._size = struct.unpack("<I", data[4:8])[0]
            if len(data) >= 16:
                self._cmd, self._checksum, self._secret_key, self._serial = struct.unpack("<HHHH", data[8:16])
            if len(data) > 16:
                self._payload = data[16:]

    def is_valid(self, check_sum=True, check_size=False):
        ret = self._header == b"\x50\x50\x82\x7D"
        if check_size:
            ret &= self._size == len(self._payload) + 8
        if check_sum:
            data = bytes(self)
            ret &= self._checksum == struct.unpack("<H", data[10:12])[0]
        return ret

    @property
    def secret_key(self):
        return self._secret_key

    @property
    def size(self):
        return self._size

    @property
    def serial(self):
        return self._serial

    @property
    def cmd(self):
        return self._cmd

    @property
    def payload(self):
        return self._payload

    def __bytes__(self):
        self._size = len(self.payload or b'') + 8
        output = self._header
        checksum = 0
        output += struct.pack("<IHHHH", self._size, self._cmd, checksum, self._secret_key, self._serial)
        if self.payload:
            output += self.payload
        checksum = self.calculate_checksum(output)
        return self._header + struct.pack("<IHHHH", self._size, self._cmd, checksum,
                                          self._secret_key, self._serial) + self.payload

    @staticmethod
    def calculate_checksum(input_list):
        _sum = 0
        size = (len(input_list) // 2) * 2
        for x in range(8, size, 2):
            if x == 10:
                continue
            _sum += struct.unpack("<H", input_list[x:x+2])[0]
            if (_sum >> 31) & 1 == 1:
                _sum = (_sum & 0xFFFF) + (_sum >> 0x10)
        if len(input_list) % 2 > 0:
            _sum += input_list[-1]
        num = _sum
        while num > 0:
            num = _sum >> 0x10
            _sum = (_sum & 0xFFFF) + num
        return (~_sum) & 0xFFFF


class DataBuffer(object):
    def __init__(self, encoding: str, raw_buffer: bytes = None):
        self._data = raw_buffer or b''
        self._encoding = encoding

    def add_leading_len(self, size=4):
        if size == 2:
            fmt = "<H"
        elif size == 1:
            fmt = "<B"
        else:
            fmt = "<I"
        self._data = struct.pack(fmt, len(self._data)) + self._data
    
    def add_leading(self, val, size=4):
        if size == 2:
            fmt = "<H"
        elif size == 1:
            fmt = "<B"
        else:
            fmt = "<I"
        self._data = struct.pack(fmt, val) + self._data

    def remove_leading(self, size=4):
        if len(self._data) > size:
            self._data = self._data[size:]

    def __len__(self):
        """
        this is different from len(bytes(self)) in that
        this one return the length of the internal data
        but the other one return data length + 4
        """
        return len(self._data)

    @property
    def data(self):
        return self._data
        # return self._data[self._start:] if len(self._data) > self._start else []
        # return self._data[self._input_leading:] if len(self._data) > self._input_leading else []

    def segment(self, offset, size):
        return self._data[offset: offset + size]
        # return bytes(self)[offset: offset + size]
    # @property
    # def attLogs(self):
    #     if (len(self) % 40) > 0:
    #         return None
    #     return [AttLog.from_bytes(x) for x in split_list(self.data, len(self) // 40)]

    @property
    def op_logs(self):
        if len(self) % 16 > 0:
            return None
        return [OpLog.from_bytes(x) for x in split_list(self.data, len(self) // 16)]

    @property
    def att_logs(self):
        if len(self) % 40 > 0:
            return None
        return [AttLog.from_bytes(x, self._encoding) for x in split_list(self.data, len(self) // 40)]

    @property
    def users(self):
        if len(self.data) % 72 > 0:
            return None
        return [UserInfo.from_bytes(x, self._encoding) for x in split_list(self.data, len(self) // 72)]
    
    @property
    def fps(self):
        index = 0
        ret = []
        while index < len(self.data):
            seg_len = struct.unpack("<H", self.data[index: index + 2])[0]
            index += 2
            seg_len -= 2
            ret.append(FPInfo.from_bytes(self.data[index: index + seg_len]))
            index += seg_len
        return ret

    def append(self, part):
        self._data += part

    def __bytes__(self):
        return self._data
        # return struct.pack("<I", len(self.data)) + self.data

    def __hash__(self):
        a = 0
        # for x in bytes(self):
        for x in self._data:
            a = (a << 0x4) + x
            b = a & 0xF0000000
            if b > 0:
                c = b >> 0x18
                if (c & 0x80) > 0:
                    c = c | 0xFFFFFF00
                c = c ^ b
                a = a ^ c
            a = a & 0xFFFFFFFF
        b = ((((a * 0x40404081) >> 32) & 0xFFFFFFFF) >> 0x16) * 0xFEFFFF
        return a - (b & 0xFFFFFFFF) if len(self) > 0 else 0



from .devices import ZMM100_TFT, ZMM220_TFT


# from .virtual import VirtualDevice


class Client(object):
    def __init__(self, device):
        self._device = device

    def __enter__(self):
        if self._device.connect():
            return self._device
        else:
            raise Exception("connection error")
    
    def __exit__(self, _type, value, traceback):
        self._device.disconnect()


class ZMM100Client(Client):
    def __init__(self, ip, port, encoding):
        super().__init__(ZMM100_TFT(ip, port, encoding))


class ZMM220Client(Client):
    def __init__(self, ip, port, encoding):
        super().__init__(ZMM220_TFT(ip, port, encoding))

# class ServerConnection(threading.Thread):
#     def __init__(self, thread_id, conn, server):
#         threading.Thread.__init__(self)
#         self._conn = conn
#         self._threads = server.threads
#         self._remove_lock = server.thread_remove_lock
#         self._server = server
#         self._machine_id = server._machine_id
#         self._id = thread_id
#         self._buffer = None
#         self._connected = False
#         # self._stopEx = threading.Event()
#         self._logger = logging.getLogger("server")
#         self._blogger = logging.getLogger("buffer")
#         self._virtual = VirtualDevice(server._machine_id, server._database_url)
#         self._last_op = ""
#         try:
#             self._virtual.connect()
#             self._client = ClientConnection(self._virtual.host, self._virtual.port)
#         except Exception as ex:
#             self._logger.error(str(ex))
#             self._client = None
#
#     def clean(self):
#         self._remove_lock.acquire()
#         while self in self._threads:
#             self._threads.remove(self)
#         self._remove_lock.release()
#         self._logger.info("removing thread {0} and remaining thread {1}".format(self._id, len(self._threads)))
#
#     def stop(self):
#         self._connected = False
#         # self._stopEx.set()
#
#     def stopped(self):
#         return not self._connected # self._stopEx.isSet()
#
#     def createPck(self, copy_from, payload=b'', cmd_name="ack", secret_key=None):
#         return packet(cmd=device_cmd[cmd_name],
#                       serial=copy_from.serial,
#                       secret_key=copy_from.secret_key if secret_key is None else secret_key,
#                       payload=payload)
#
#     def _readPck(self):
#         data = b''
#         seg_len = 0x1000
#         header = self._conn.recv(8)
#         pck = packet(header)
#         if not pck.isValid(False, False):
#             raise Exception("received packet is not valid")
#         rem_size = pck.size - len(data)
#         while rem_size > 0:
#             temp_data = self._conn.recv(seg_len if rem_size > seg_len else rem_size)
#             if not temp_data:
#                 raise Exception("connection is closed by the client")
#             data += temp_data
#             rem_size = pck.size - len(data)
#         data = header + data
#         # if debug:
#         #     self._blogger.debug("REQ ({} bytes): {}".format(len(data),
#         #         str.join(' ', [hex(char) for char in data])))
#         return packet(data)
#
#     def readPcks(self):
#         pcks = []
#         pck = self._readPck()
#         pcks.append(pck)
#         # if pck.cmd == device_cmd["recv_buff_header"]:
#         #     pck = self._readPck()
#         #     pcks.append(pck)
#         # while pck.cmd == device_cmd["recv_buff_content"]:
#         #     pck = self._readPck()
#         #     pcks.append(pck)
#         # pcks.append(self._readPck())
#         return pcks
#
#
#     def run(self):
#         # att_offset = 0
#         # att_size = 0
#         idle = 0
#         # connect to machine
#         self._connected = self._client and self._client.connect()
#         try:
#             while self._connected:
#                 command_processed = False
#                 ready = select([self._conn], [], [], 3)
#                 # data = b''
#                 if not ready[0]:
#                     idle += 3
#                     # print(idle)
#                 else:
#                     idle = 0
#                     pcks = self.readPcks()
#                     # print("{}: {}".format(hex(pcks[0].cmd), len(pcks)))
#                     pck = pcks[0]
#
#                     if pck.cmd == device_cmd["recv_buff_header"]:
#                         # print(dump(pck.payload))
#                         self._buffer = DataBuffer()
#                     elif pck.cmd == device_cmd["recv_buff_content"]:
#                         self._buffer.append(pck.payload)
#                     # elif pck.cmd == device_cmd["check_hash"]:
#                         # if not self._buffer:
#                         #     recv = bytes(self.createPck(pck, struct.pack("<I", 0)))
#                         # else:
#                         #     recv = bytes(self.createPck(pck, struct.pack("<I", hash(self._buffer))))
#                         # command_processed = False
#                     elif pck.cmd == device_cmd["del_face"]:
#                         person_id = get_null_term_str(pck.payload, 0, 24)
#                         face_id = pck.payload[25]
#                         self._virtual.delFP(person_id, face_id)
#                     elif pck.cmd == device_cmd["set_face"]:
#                         person_id = get_null_term_str(pck.payload, 0, 24)
#                         face_index, buffer_len = struct.unpack("<BH", pck.payload[25:28])
#                         if buffer_len == len(self._buffer.data):
#                             self._virtual.setUserFace(person_id, self._buffer.data, face_index)
#
# elif pck.cmd == device_cmd["get_face"]: person_id = get_null_term_str(pck.payload, 0, 24) face_id = pck.payload[25]
# face_id = struct.unpack("<BB", pck.payload[24:26])[0] data = self._virtual.getUserFace(person_id, face_id) if not
# data: recv = bytes(self.createPck(pck, cmd_name="no_data")) else: payload = struct.pack("<II", len(data),
# 0x10000) recv = bytes(self.createPck(pck, payload=payload, cmd_name="recv_buff_header")) recv += bytes(
# self.createPck(pck, payload=data, cmd_name="recv_buff_content",secret_key=0)) recv += bytes(self.createPck(pck))
# command_processed = True elif pck.cmd == device_cmd["get_fp_ex"]: user_id, finger_id = struct.unpack("<HB",
# pck.payload[:3]) fp = self._virtual.getFP(user_id, finger_id) # print("fp len: {}".format(len(fp.data))) if not fp:
# recv = bytes(self.createPck(pck, cmd_name="no_data")) else: data = fp.data + struct.pack("<6xB", fp.enabled)
# payload = struct.pack("<II", len(data), 0x10000) recv = bytes(self.createPck(pck, payload=payload,
# cmd_name="recv_buff_header")) recv += bytes(self.createPck(pck, payload=data, cmd_name="recv_buff_content",
# secret_key=0)) recv += bytes(self.createPck(pck)) command_processed = True elif pck.cmd == device_cmd["set_fp_ex"]:
# fp_info = models.FPInfo() fp_info.user_id, fp_info.finger_id, fp_info.enabled, buffer_len = struct.unpack("<HBBH",
# pck.payload) fp_info.data = self._buffer.data # update fp in virtual device self._virtual.setFP(fp_info) #
# self._last_op = "set_fp_ex" # recv = bytes(self.createPck(pck)) command_processed = False elif pck.cmd ==
# device_cmd["del_fp_ex"]: # delete fp in virtual device person_id = get_null_term_str(pck.payload, 0, 24) finger_id =
# pck.payload[24] self._virtual.delFP(person_id, finger_id)
#
#                     elif pck.cmd == device_cmd["set_user"]:
#                         user_info = models.UserInfo.from_bytes(pck.payload)
#                         self._virtual.setUser(user_info)
#                         # print(dump(pck.payload))
#                         # recv = bytes(self.createPck(pck))
#                         # command_processed = True
#                     # elif pck.cmd == device_cmd["save_data"]:
#                     #     if self._last_op:
#                     #         recv = bytes(self.createPck(pck))
#                     #         command_processed = True
#                     #         self._last_op = ""
#
# elif pck.cmd == device_cmd["disconnect"]: self._connected = False elif pck.cmd == device_cmd["logs_count"]: # st =
# self._db_session.query(models.MachineState).filter_by(machine_id=self._machine_id).first() recv = bytes(
# self.createPck(pck, bytes(self._virtual.getState()))) command_processed = True
#
#                     elif pck.cmd == device_cmd["del_logs"]:
#                         self._virtual.delAttLogs()
#                         recv = bytes(self.createPck(pck))
#                         command_processed = True
#
#                     elif pck.cmd == device_cmd["start_buff_stream"]:
#                         buffer_type = struct.unpack("<I", pck.payload[:4])[0]
#                         self._buffer = DataBuffer()
#                         # buffer_data = b''
#                         if buffer_type == 0xD01:
#                             # att_log_set = 0
#                             # att_size = 0
#                             logs = self._virtual.getAttLogs()
#                             for log in logs:
#                                 self._buffer.append(bytes(log)) # += bytes(log)
#                             self._last_op = "true"
#                         elif buffer_type == 0x05000901:
#                             users = self._virtual.getUsers()
#                             for user in users:
#                                 self._buffer.append(bytes(user)) # += bytes(user)
#                             self._last_op = "true"
#                         elif buffer_type == 0x02000701:
#                             fps = self._virtual.getFPS()
#                             for fp in fps:
#                                 temp = bytes(fp)
#                                 self._buffer.append(struct.pack("<H", len(temp) + 2) + temp)
#                             self._last_op = "true"
#
#                         else:
#                             self._buffer = None
#                             self._last_op = ""
#                         if not self._buffer or not self._buffer.data:
#                             recv = bytes(self.createPck(pck, cmd_name="no_data"))
#                         else:
#                             # self._buffer = DataBuffer(buffer_data)
#                             self._buffer.add_leading_len()
#                             # self._buffer = DataBuffer(struct.pack("<I", len(buffer_data)) + buffer_data)
#                             buffer_len = struct.pack("<I", len(self._buffer))
#                             payload = b'\x00' + buffer_len + buffer_len + struct.pack("<I", hash(self._buffer))
#                             recv = bytes(self.createPck(pck, payload))
#                             command_processed = True
#
#                     elif pck.cmd == device_cmd["buff_stream"] and self._buffer is not None:
#                         offset, size = struct.unpack("<II", pck.payload[:8])
#                         # size = struct.unpack("<I", pck.payload[4:8])[0]
#                         # payload = struct.pack("<II", size, 0x10000) # pck.payload[4:] + struct.pack("<I", 0x10000)
#                         recv = bytes(self.createPck(pck, struct.pack("<II", size, 0x10000), "recv_buff_header"))
#
#                         # self._conn.sendall(recv)
#                         # payload = bytes(self._buffer)[att_offset: att_size]
#                         recv += bytes(self.createPck(pck, self._buffer.segment(offset, size), "recv_buff_content", 0))
#                         # recv += bytes(packet(cmd=device_cmd[],
#                         #                     serial=pck.serial,
#                         #                     secret_key=0, payload=))
#                         recv += bytes(self.createPck(pck))
#                         command_processed = True
#                     elif pck.cmd == device_cmd["end_buff_stream"] and self._buffer is not None:
#                         if self._last_op:
#                             self._last_op = ""
#                             recv = bytes(self.createPck(pck))
#                             command_processed = True
#                         self._buffer = None
#
#                     # elif pck.cmd == device_cmd["connect"]:
#                     #     recv = bytes(self.createPck(pck, cmd_name="accept_conn"))
#                     elif pck.cmd == device_cmd["soft_ver"]:
#                         recv = bytes(self.createPck(pck, self._virtual.SoftwareVersion.encode('latin-1'), ))
#                         command_processed = True
#                     elif pck.cmd == device_cmd["del_user_pic"]:
#                         person_id = pck.payload.decode('cp1256').strip('.jpg\x00')
#                         self._virtual.delUserPic(person_id)
#                     elif pck.cmd == device_cmd["set_user_pic"]:
#                         person_id = get_null_term_str(pck.payload, 0, -1).strip(".jpg")
#                         self._virtual.setUserPic(person_id, self._buffer.data)
#                     elif pck.cmd == device_cmd["get_user_pic"]:
#                         person_id = pck.payload.decode('cp1256').strip('.jpg\x00')
#                         pic = self._virtual.getUserPic(person_id)
#                         if not pic:
#                             recv = bytes(self.createPck(pck, cmd_name="no_pic"))
#                         else:
#                             payload = struct.pack("<II", len(pic), 0xFFF8)
#                             recv = bytes(self.createPck(pck, payload, "recv_buff_header"))
#                             recv += bytes(self.createPck(pck, pic, "recv_buff_content", 0))
#                             recv += bytes(self.createPck(pck))
#                         command_processed = True
#                     elif pck.cmd == device_cmd["get_table_struct"]:
#                         temp = self._virtual.DatabaseStructure.encode('latin-1')
#                         payload = struct.pack("<II", len(temp), 0xFFF8)
#                         recv = bytes(self.createPck(pck, payload, "recv_buff_header"))
#                         recv += bytes(self.createPck(pck, temp, "recv_buff_content", 0))
#                         recv += bytes(self.createPck(pck))
#                         command_processed = True
#
# elif pck.cmd == device_cmd["get_data"]: option_name = pck.payload.decode('latin-1').strip(' \x00') opt =
# self._virtual.getSysOption(option_name) if not opt: recv = bytes(self.createPck(pck, cmd_name="no_sys_op")) else:
# recv = bytes(self.createPck(pck, "{}={}\x00".format(opt.prop_name, opt.prop_val).encode('latin-1')))
# command_processed = True
#
#                     elif pck.cmd == device_cmd["query_sys_op"]:
#                         query_string = pck.payload.decode('latin-1')
#                         output_string = self._virtual.querySysOption(query_string)
#                         recv = bytes(self.createPck(pck, output_string.encode('latin-1')))
#                         command_processed = True
#
#                     # elif pck.cmd == device_cmd["ex_op_log_recv"]:
#                     #     recv = bytes(self.createPck(pck, cmd_name="ex_op_log_recv"))
#                     #     command_processed = True
#
#                     elif pck.cmd == device_cmd["get_pin_width"]:
#                         recv = bytes(self.createPck(pck, struct.pack("<B", 9)))
#                         command_processed = True
#                     # else:
#                     #     recv = bytes(self.createPck(pck))
#                     #     command_processed = True
#                     # default behaviour is to send the request to the machine
#                     # and bypass response to the client
#                     if not command_processed:
#                         # print("send to machine")
#                         self._client._request = b''
#                         for pck in pcks:
#                             self._client._request += bytes(pck)
#                         # self._client._request = data
#                         self._client.send()
#                         self._client.receive()
#                         recv = bytes(self._client._response)
#                         if self._client._response.cmd == device_cmd["recv_buff_header"]:
#                             self._client.receive()
#                             recv += bytes(self._client._response)
#                             self._client.receive()
#                             recv += bytes(self._client._response)
#
#                         # recv = self._client.generic_receive()
#                     if len(recv) > 0:
#                         # if debug:
#                         #     self._blogger.debug("RES ({} bytes): {}".format(len(recv),
#                         #         str.join(' ', [hex(char) for char in recv])))
#                         self._conn.sendall(recv)
#                     # connected = not self.stopped()
#                 if idle >= 1800:
#                     self._logger.info("idle reach limit for thread {}".format(self._id))
#                     self._connected = False
#
#                 # if not connected or self.stopped():
#                 #     break
#         except Exception as ex:
#             self._logger.error(str(ex))
#         finally:
#             self._conn.close()
#             # close machine socket
#             if self._client:
#                 self._client.disconnect()
#             self.clean()

                
# class Server(object):
#     def __init__(self, machine_id, host="0.0.0.0", port=4370, database_url=""):
#         self._machine_id = machine_id
#         self._host = host
#         self._port = port
#         self._socket = None
#         self.threads = []
#         self.thread_remove_lock = threading.Lock()
#         self._logger = logging.getLogger("server")
#         self._database_url = database_url
#
#     def cleanup(self):
#         self._socket.close()
#         for th in self.threads:
#             th.stop()
#         while len(self.threads) > 0:
#             self.threads[0].join()
#
#     def create_thread_id(self):
#         id = randint(0x1000, 0xFFFF)
#         while id in [x._id for x in self.threads]:
#             id = randint(0x1000, 0xFFFF)
#         return id
#
#     def run(self):
#         self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
#         self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
#         self._socket.bind((self._host, self._port))
#         self._socket.listen(10)
#         try:
#             if not self._database_url:
#                 raise Exception("database_url cannot be empty")
#             while True:
#                 conn, addr = self._socket.accept()
#                 self._logger.info("machine with ip ({}) is connected to system".format(addr))
#                 th_id = self.create_thread_id()
#                 th = ServerConnection(th_id, conn, self)
#                 th.start()
#                 self.threads.append(th)
#                 self._logger.info("thread {} added with total thread {}".format(th_id, len(self.threads)))
#         except Exception as ex:
#             self._logger.error(str(ex))
#         finally:
#             self.cleanup()



import os
import logging
import logging.handlers

debug = os.environ.get("_DEBUG")
def setup_logger():
    log_path = os.path.join(os.environ.get("LOG_PATH") or "", "server.log")
    handler = logging.handlers.RotatingFileHandler(log_path, maxBytes=1*1024*1024, backupCount=5)
    fmt = logging.Formatter('%(levelname)s(%(asctime)s): %(message)s')
    handler.setFormatter(fmt)
    logger = logging.getLogger("server")
    logger.setLevel(logging.INFO)
    logger.addHandler(handler)

    if debug:
        log_path = os.path.join(os.environ.get("LOG_PATH") or "", "buffer.log")
        handler = logging.handlers.RotatingFileHandler(log_path, maxBytes=5*1024*1024, backupCount=5)
        handler.setFormatter(fmt)
        logger = logging.getLogger("buffer")
        logger.setLevel(logging.DEBUG)
        logger.addHandler(handler)

setup_logger()